// Web3DView.cpp : implementation of the CWeb3DView class
//

#include "stdafx.h"
#include "Web3D.h"

#include "Web3DDoc.h"
#include "Web3DView.h"

#include "src/w3Dcore.h"


w3DSystem			sys;
w3DCamera			cam(0, 500, 7000, 0, 0, 0);
w3DRotate			rot_yaw(0, 0, 1, 0);
w3DRotate			rot_pitch(0, 1, 0, 0);
w3DTranslate		trans(0, 0, 0);
w3DGroup			attr;
w3DModelRenderer	mod;

BOOL bLeftPushed = FALSE;
BOOL bRightPushed = FALSE;

float	cur_yaw = 0.0f;
float	cur_pitch = 0.0f;
float	cur_trs_x = 0.0f;
float	cur_trs_y = 0.0f;
float	cur_trs_z = 0.0f;
CPoint	point_last;


#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

/////////////////////////////////////////////////////////////////////////////
// CWeb3DView

IMPLEMENT_DYNCREATE(CWeb3DView, CView)

BEGIN_MESSAGE_MAP(CWeb3DView, CView)
	//{{AFX_MSG_MAP(CWeb3DView)
	ON_WM_CREATE()
	ON_WM_DESTROY()
	ON_WM_ERASEBKGND()
	ON_WM_SIZE()
	ON_WM_LBUTTONDOWN()
	ON_WM_LBUTTONUP()
	ON_WM_MOUSEMOVE()
	ON_WM_RBUTTONDOWN()
	ON_WM_RBUTTONUP()
	ON_WM_LBUTTONDBLCLK()
	//}}AFX_MSG_MAP
	// Standard printing commands
	ON_COMMAND(ID_FILE_PRINT, CView::OnFilePrint)
	ON_COMMAND(ID_FILE_PRINT_DIRECT, CView::OnFilePrint)
	ON_COMMAND(ID_FILE_PRINT_PREVIEW, CView::OnFilePrintPreview)
END_MESSAGE_MAP()

/////////////////////////////////////////////////////////////////////////////
// CWeb3DView construction/destruction

CWeb3DView::CWeb3DView()
{
	// TODO: add construction code here

}

CWeb3DView::~CWeb3DView()
{
}

BOOL CWeb3DView::PreCreateWindow(CREATESTRUCT& cs)
{
	// TODO: Modify the Window class or styles here by modifying
	//  the CREATESTRUCT cs

	return CView::PreCreateWindow(cs);
}

/////////////////////////////////////////////////////////////////////////////
// CWeb3DView drawing

void CWeb3DView::OnDraw(CDC* pDC)
{
	CWeb3DDoc* pDoc = GetDocument();
	ASSERT_VALID(pDoc);
	// TODO: add draw code for native data here

	sys.Process();
	w3DFlushAndSwap();
}

/////////////////////////////////////////////////////////////////////////////
// CWeb3DView printing

BOOL CWeb3DView::OnPreparePrinting(CPrintInfo* pInfo)
{
	// default preparation
	return DoPreparePrinting(pInfo);
}

void CWeb3DView::OnBeginPrinting(CDC* /*pDC*/, CPrintInfo* /*pInfo*/)
{
	// TODO: add extra initialization before printing
}

void CWeb3DView::OnEndPrinting(CDC* /*pDC*/, CPrintInfo* /*pInfo*/)
{
	// TODO: add cleanup after printing
}

/////////////////////////////////////////////////////////////////////////////
// CWeb3DView diagnostics

#ifdef _DEBUG
void CWeb3DView::AssertValid() const
{
	CView::AssertValid();
}

void CWeb3DView::Dump(CDumpContext& dc) const
{
	CView::Dump(dc);
}

CWeb3DDoc* CWeb3DView::GetDocument() // non-debug version is inline
{
	ASSERT(m_pDocument->IsKindOf(RUNTIME_CLASS(CWeb3DDoc)));
	return (CWeb3DDoc*)m_pDocument;
}
#endif //_DEBUG

/////////////////////////////////////////////////////////////////////////////
// CWeb3DView message handlers

int CWeb3DView::OnCreate(LPCREATESTRUCT lpCreateStruct) 
{
	if (CView::OnCreate(lpCreateStruct) == -1)
		return -1;
	
	sys.AdjustCSInfo(*lpCreateStruct);
	
	return 0;
}

void CWeb3DView::OnDestroy() 
{
	CView::OnDestroy();
	
	sys.Stop();
}

BOOL CWeb3DView::OnEraseBkgnd(CDC* pDC) 
{
	return FALSE;
}

void CWeb3DView::OnSize(UINT nType, int cx, int cy) 
{
	CView::OnSize(nType, cx, cy);
	
	sys.Resize(cx, cy);
	Invalidate();
}

void CWeb3DView::OnInitialUpdate() 
{
	CView::OnInitialUpdate();
	
	// set active camera
	sys.SetActiveCamera(&cam);
	sys.Init(m_hWnd);

	mod.Init("out.bin");

	attr.Add(&trans);
	attr.Add(&rot_pitch);
	attr.Add(&rot_yaw);
	mod.SetAttrib(&attr);

	// register the scene
	sys.SetModeler(&mod);
	sys.Resume();
}

void CWeb3DView::OnLButtonDblClk(UINT nFlags, CPoint point) 
{
	unsigned long id = sys.PickObject(point.x, point.y);

	if (id > 0)
	{
		STRUCT_OBJECT*	pTarget;
		mod.QueryObject(id, (void**)&pTarget);
		w3DWinMsg("name: %s\n", pTarget->name);
	}
	
	CView::OnLButtonDblClk(nFlags, point);
}

void CWeb3DView::OnLButtonDown(UINT nFlags, CPoint point) 
{
	SetCapture();

	bLeftPushed = TRUE;

	CView::OnLButtonDown(nFlags, point);
}

void CWeb3DView::OnLButtonUp(UINT nFlags, CPoint point) 
{
	ReleaseCapture();

	bLeftPushed = FALSE;
	
	CView::OnLButtonUp(nFlags, point);
}

void CWeb3DView::OnMouseMove(UINT nFlags, CPoint point) 
{
	if (bLeftPushed && bRightPushed)
	{
		// It means 'PANNING' operation.
		cur_trs_x += (float)(point.x - point_last.x);
		cur_trs_y -= (float)(point.y - point_last.y);
		trans.Set(cur_trs_x, cur_trs_y, cur_trs_z);

		Invalidate();
	}
	else if (bLeftPushed)
	{
		// It means 'ROTATION' operation.
		cur_yaw += (float)(point.x - point_last.x);
		rot_yaw.SetAngle(cur_yaw);

		cur_pitch += (float)(point.y - point_last.y);
		if (cur_pitch > 90.0f)
			cur_pitch = 90.0f;
		if (cur_pitch < -90.0f)
			cur_pitch = -90.0f;

		rot_pitch.SetAngle(cur_pitch);

		Invalidate();
	}
	else if (bRightPushed)
	{
		// It means 'ZOOM' operation.
		cur_trs_z += (float)(point.y - point_last.y);
		trans.Set(cur_trs_x, cur_trs_y, cur_trs_z);

		Invalidate();
	}

	point_last = point;
	
	CView::OnMouseMove(nFlags, point);
}

void CWeb3DView::OnRButtonDown(UINT nFlags, CPoint point) 
{
	SetCapture();

	bRightPushed = TRUE;
	
	CView::OnRButtonDown(nFlags, point);
}

void CWeb3DView::OnRButtonUp(UINT nFlags, CPoint point) 
{
	ReleaseCapture();

	bRightPushed = FALSE;
	
	CView::OnRButtonUp(nFlags, point);
}

