

/*
**	W3D basic.  Jinhong Kim, cheriz@empal.com
**
**  THIS IS COMPLETELY FREE!  ANYONE WILL BE PERMITTED TO MODIFY THE CODE FOR HIS/HER OWN PURPOSE.
**  (THE PURPOSE INCLUDES THE COMMERCIAL)
**  BUT, YOU SHOULD KNOW THAT I'M FREE FROM ANY-RESPONSIBILITIES FOR- 
**  ANY SYSTEM RELATED ERRORS MAY OCCUR WHILE USING THE CODE.
**  FREE YOUR MIND.
*/



#include <windows.h>
#include <gl/gl.h>
#include <gl/glu.h>
#include <gl/glaux.h>
#include <math.h>
#include <time.h>
#include <stdio.h>
#include <float.h>

#include "dbllist.h"
#include "w3Dcore.h"
#include "w3dmath.h"


#pragma warning (disable : 4710)	// not inlined warning; This will keep on.



//////////////////////////////////////////////////////////////////////////////
// INTERNAL ONLY FUNCTION PROTOTYPE FORWARD DECLARATIONS

static BOOL classify_attrib(w3DObject *pObj);


char chkPowerOf2(int n)
{
	int p2 = 2;

	while (1)
	{
		if (p2 == n) return 1;

		p2 *= 2;

		if (p2 > n) break;
	}

	return 0;
}


int getNearestPowerOf2(int n)
{
	int p2 = 2;
	int last;
	int d1, d2;


	while (1)
	{
		if (p2 == n) return n;

		last = p2;

		p2 *= 2;

		if (p2 > n)	break;
	}

	d1 = n - last;
	d2 = p2 - n;

	if (d1 < d2) return last;

	return p2;
}


int getUpperPowerOf2(int n)
{
	int p2 = 2;

	while (1)
	{
		if (p2 == n) return n;

		p2 *= 2;

		if (p2 > n)	break;
	}

	return p2;
}


int getLowerPowerOf2(int n)
{
	int p2 = 2;
	int last;

	while (1)
	{
		if (p2 == n) return n;

		last = p2;
		p2 *= 2;

		if (p2 > n)	break;
	}

	return last;
}


void ReduceToUnitf(float vector[3])
{
	float length;
	
	// Calculate the length of the vector		
	length = (float)sqrt((vector[0]*vector[0]) + 
					(vector[1]*vector[1]) +
					(vector[2]*vector[2]));

	// Keep the program from blowing up by providing an exceptable
	// value for vectors that may calculated too close to zero.
	if(length == 0.0f)
		length = 1.0f;

	// Dividing each element by the length will result in a
	// unit normal vector.
	vector[0] /= length;
	vector[1] /= length;
	vector[2] /= length;
}

// Calculate the normal vector in a reduced form(Unit Vector)
void calcNormalf(float v[3][3], float out[3])
{
	float v1[3], v2[3];
	static const int x = 0;
	static const int y = 1;
	static const int z = 2;

	// Calculate two vectors from the three points
	v1[x] = v[0][x] - v[1][x];
	v1[y] = v[0][y] - v[1][y];
	v1[z] = v[0][z] - v[1][z];

	v2[x] = v[1][x] - v[2][x];
	v2[y] = v[1][y] - v[2][y];
	v2[z] = v[1][z] - v[2][z];

	// Take the cross product of the two vectors to get
	// the normal vector which will be stored in out
	out[x] = v1[y]*v2[z] - v1[z]*v2[y];
	out[y] = v1[z]*v2[x] - v1[x]*v2[z];
	out[z] = v1[x]*v2[y] - v1[y]*v2[x];
}



//////////////////////////////////////////////////////////////////////////////
// IMPLEMENTATION: w3DObject

w3DObject::w3DObject()
{
}

w3DObject::~w3DObject()
{
}


//////////////////////////////////////////////////////////////////////////////
// IMPLEMENTATION: w3DGroup

w3DGroup::w3DGroup() 
{
}

w3DGroup::~w3DGroup() 
{
	Empty();
}

W3DGROUPNODETYPE* w3DGroup::Add(w3DObject* item)
{
	return list.PushBack(item);
}

void w3DGroup::FindNext()
{
	node++;
}

void w3DGroup::FindFirst()
{
	node = list.Front();
}

w3DObject* w3DGroup::GetData()
{
	return *node;
}

W3DGROUPNODETYPE& w3DGroup::GetCurNode()
{
	return node;
}

void w3DGroup::SetCurNode(W3DGROUPNODETYPE* pnode)
{
	node = pnode;
}

BOOL w3DGroup::DeleteFirst()
{
	BOOL b;
	b = list.RemoveFront();
	FindFirst();
	return b;
}

void w3DGroup::Empty()
{
	list.RemoveAll();
}

void w3DGroup::SetParent(w3DGroup* pParent_)
{
	pParent = pParent_;
}

w3DGroup* w3DGroup::GetParent()
{
	return pParent;
}

int w3DGroup::GetTotalNum()
{
	return list.GetTotalNum();
}


//////////////////////////////////////////////////////////////////////////////
// IMPLEMENTATION: w3DModel

w3DModelBase::w3DModelBase()
{
	attr = NULL;
}

w3DModelBase::~w3DModelBase() 
{
}

w3DGroup* w3DModelBase::GetAttrib()
{
	return attr;
}

void w3DModelBase::SetAttrib(w3DGroup* attr)
{
	this->attr = attr;
}



//////////////////////////////////////////////////////////////////////////////
// IMPLEMENTATION: w3DSystem

const GLfloat w3DSystem::PI				= 3.14159265359f;
const GLfloat w3DSystem::TORAD			= 0.017453292519944f;
const GLfloat w3DSystem::TODEG			= 57.295780490443f;
HWND	w3DSystem::hWnd					= NULL;
HDC		w3DSystem::hDC					= NULL;
HGLRC	w3DSystem::hRC					= NULL;
BOOL	w3DSystem::bInitialized			= FALSE;
float	w3DSystem::Aspect				= 0.0f;
GLfloat w3DSystem::bkColor[4]			= {0.0f, 0.0f, 0.0f, 1.0f};
BOOL	w3DSystem::cull_face			= TRUE;
BOOL	w3DSystem::PolygonMode			= FALSE;
BOOL	w3DSystem::SmoothShade			= TRUE;
BOOL	w3DSystem::RenderingMode		= TRUE;
w3DModelBase*	w3DSystem::pModeler		= NULL;
w3DCamera*	w3DSystem::pCamera			= NULL;

GLfloat w3DSystem::ambientLight[4] = {0.15f, 0.15f, 0.15f, 1.0f};
GLfloat w3DSystem::diffuseLight[4] = {1.0f, 1.0f, 1.0f, 1.0f};
GLfloat w3DSystem::specularLight[4] = {1.0f, 1.0f, 1.0f, 1.0f};
GLfloat w3DSystem::lightPos[4] = {-1000.0f, 1000.0f, 1000.0f, 1.0f};

BYTE w3DSystem::DepthBufferSize = 16;
BYTE w3DSystem::ColorDepth = 32;


w3DSystem::w3DSystem()
{
	SetBackGroundColor(0.0f, 0.0f, 0.0f, 1.0f);
	SetAmbientLight(0.15f, 0.15f, 0.15f, 1.0f);
	SetDiffuseLight(1.0f, 1.0f, 1.0f, 1.0f);
	SetSpecularLight(1.0f, 1.0f, 1.0f, 1.0f);
	SetLightPos(-1000.0f, 1000.0f, 1000.0f, 1.0f);
	SetColorDepth(32);
	SetDepthBufferSize(16);
	stop = FALSE;
}

w3DSystem::~w3DSystem()
{
	wglDeleteContext(w3DSystem::hRC);
	::ReleaseDC(w3DSystem::hWnd, w3DSystem::hDC);
}

// THIS IS A STATIC MEMBER FUNCTION
void w3DSystem::SetAmbientLight(GLfloat r, GLfloat g, GLfloat b, GLfloat a)
{
	ambientLight[0] = r;
	ambientLight[1] = g;
	ambientLight[2] = b;
	ambientLight[3] = a;
}

// THIS IS A STATIC MEMBER FUNCTION
void w3DSystem::SetDiffuseLight(GLfloat r, GLfloat g, GLfloat b, GLfloat a)
{
	diffuseLight[0] = r;
	diffuseLight[1] = g;
	diffuseLight[2] = b;
	diffuseLight[3] = a;
}

// THIS IS A STATIC MEMBER FUNCTION
void w3DSystem::SetSpecularLight(GLfloat r, GLfloat g, GLfloat b, GLfloat a)
{
	specularLight[0] = r;
	specularLight[1] = g;
	specularLight[2] = b;
	specularLight[3] = a;
}

// THIS IS A STATIC MEMBER FUNCTION
void w3DSystem::SetLightPos(GLfloat x, GLfloat y, GLfloat z, GLfloat a)
{
	lightPos[0] = x;
	lightPos[1] = y;
	lightPos[2] = z;
	lightPos[3] = a;
}

// THIS IS A STATIC MEMBER FUNCTION
void w3DSystem::SetBackGroundColor(GLfloat r, GLfloat g, GLfloat b, GLfloat a)
{
	bkColor[0] = r;
	bkColor[1] = g;
	bkColor[2] = b;
	bkColor[3] = a;
}

// THIS IS A STATIC MEMBER FUNCTION
void w3DSystem::SetCullFace(BOOL bi)
{
	cull_face = bi;
}

// A STATIC MEMBER FUNCTION
void w3DSystem::SetPolygonMode(BOOL mode)
{
	PolygonMode = mode;
}

void w3DSystem::AdjustCSInfo(CREATESTRUCT &cs)
{
	cs.style |= (WS_CLIPCHILDREN | WS_CLIPSIBLINGS | CS_OWNDC);
}

void w3DSystem::SetupRC()
{
	if (cull_face)
	{
		glEnable(GL_CULL_FACE);
		glCullFace(GL_BACK);
	}
	else
		glDisable(GL_CULL_FACE);

	glEnable(GL_DEPTH_TEST);
	glFrontFace(GL_CCW);

	if (PolygonMode == 0)
		glPolygonMode(GL_FRONT, GL_FILL);
	else
		glPolygonMode(GL_FRONT, GL_LINE);

	glShadeModel(GL_SMOOTH);
	glEdgeFlag(TRUE);


//	glHint(GL_LINE_SMOOTH_HINT, GL_NICEST);
//	glHint(GL_POINT_SMOOTH_HINT, GL_NICEST);
//	glHint(GL_POLYGON_SMOOTH_HINT, GL_NICEST);
//	glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST);

//	glEnable(GL_LINE_SMOOTH);
//	glEnable(GL_POINT_SMOOTH);
//	glEnable(GL_POLYGON_SMOOTH);


	// light setting
	glEnable(GL_LIGHTING);

	glLightModelfv(GL_LIGHT_MODEL_AMBIENT, ambientLight);
//	GLfloat lv = 0.0f;
//	glLightModelfv(GL_LIGHT_MODEL_LOCAL_VIEWER, &lv);
//	glLightModeli(GL_LIGHT_MODEL_TWO_SIDE, GL_TRUE);
	glLightfv(GL_LIGHT0, GL_AMBIENT, ambientLight);
	glLightfv(GL_LIGHT0, GL_DIFFUSE, diffuseLight);
	glLightfv(GL_LIGHT0, GL_SPECULAR, specularLight);
	glEnable(GL_LIGHT0);
	
	glEnable(GL_COLOR_MATERIAL);
	glColorMaterial(GL_FRONT, GL_AMBIENT_AND_DIFFUSE);

	GLfloat specref[] = {1.0f, 1.0f, 1.0f, 1.0f};
	glMaterialfv(GL_FRONT, GL_SPECULAR, specref);
	glMateriali(GL_FRONT, GL_SHININESS, 128);

	glClearColor(bkColor[0], bkColor[1], bkColor[2], 1.0f);
}

// STATIC FUNCTION
void w3DSystem::SetColorDepth(BYTE cd)
{
	ColorDepth = cd;
}

// STATIC FUNCTION
void w3DSystem::SetDepthBufferSize(BYTE db)
{
	DepthBufferSize = db;
}

void w3DSystem::Init(HWND hWnd)
{
	LONG nPixelFormat;

	w3DSystem::hWnd = hWnd;
	w3DSystem::hDC = ::GetDC(hWnd);

	if (NULL == w3DSystem::hWnd || NULL == w3DSystem::hDC)
		return;


	static PIXELFORMATDESCRIPTOR pfd = {
		sizeof(PIXELFORMATDESCRIPTOR),	// Size of this structure
		1,								// Version of this structure	
		PFD_DRAW_TO_WINDOW |			// Draw to Window (not to bitmap)
		PFD_SUPPORT_OPENGL |			// Support OpenGL calls in window
		PFD_DOUBLEBUFFER,				// Double buffered mode
		PFD_TYPE_RGBA,					// RGBA Color mode
		ColorDepth,						// Color depth
		0,0,0,0,0,0,					// Not used to select mode
		0,0,							// Not used to select mode
		0,								// Accumulation Buffer
		0,
		0,
		0,
		0,
		DepthBufferSize,				// Size of depth buffer
		0,								// Stencil
		0,								// Not used to select mode
		PFD_MAIN_PLANE,					// Draw in main plane
		0,								// Not used to select mode
		0,0,0 };						// Not used to select mode

	// Choose a pixel format that best matches that described in pfd
	nPixelFormat = ChoosePixelFormat(w3DSystem::hDC, &pfd);

	// Set the pixel format for the device context
	if (!SetPixelFormat(hDC, nPixelFormat, &pfd))
	{
		w3DWinMsg("OpenGL PixelFormat could not be set!\n");
		return;
	}

	// Create the rendering context
	w3DSystem::hRC = wglCreateContext(w3DSystem::hDC);
	w3DSystem::bInitialized = TRUE;

	w3DContext(TRUE);

	// setup rendering context
	SetupRC();

	// display GL informations
#ifdef W3DDEBUG
	w3DErrorTest("After SetupRC()");

	w3DDebugMsg("--- W3D Informations ---\n");
	w3DDebugMsg("System Version: %s\n", Version);
	w3DDebugMsg("--- OpenGL Informations ---\n");
	w3DDebugMsg("Vendor: %s\n", glGetString(GL_VENDOR));
	w3DDebugMsg("Renderer: %s\n", glGetString(GL_RENDERER));
	w3DDebugMsg("Version: %s\n", glGetString(GL_VERSION));

	LONG i, slen;
	const unsigned char *sptr = glGetString(GL_EXTENSIONS);
	char buffer[2500];

	slen = strlen((const char*)sptr);
	CopyMemory((void*)buffer, (void*)sptr, slen);

	for (i=0; i<slen; i++)
		if (buffer[i] == ' ')
			buffer[i] = '\n';

	w3DDebugMsg("Extensions:\n");
	w3DDebugMsg("%s\n\n", buffer);
	w3DDebugMsg("\n");
#endif

	w3DContext(FALSE);

	// do Resize first
	RECT r;
	::GetClientRect(hWnd, &r);
	Resize(r.right, r.bottom);
}


void w3DSystem::Resize(GLsizei w, GLsizei h)
{
	if (!w3DSystem::bInitialized)
		return;

	if(0 == h)
		h = 1;

	w3DContext(TRUE);

	// Set Viewport to window dimensions
    glViewport(0, 0, w, h);

	// Reset coordinate system
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();

	w3DSystem::Aspect = (GLfloat)w/(GLfloat)h;
	gluPerspective(pCamera->GetFOV(), Aspect, pCamera->GetNearDist(), pCamera->GetFarDist());

	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();

	w3DContext(FALSE);
}

void w3DSystem::Process()
{
	if (!w3DSystem::bInitialized || stop)
		return;

	if (w3DSystem::RenderingMode)
	{
		w3DContext(TRUE);
		
		glMatrixMode(GL_MODELVIEW);
		glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
		glLoadIdentity();
		
		// positioning the light
		glLightfv(GL_LIGHT0, GL_POSITION, lightPos);
		
		// apply camera transform
		pCamera->Apply();
		
		// render the scene
		w3DApplyAttrib(pModeler->GetAttrib());
		pModeler->Render();
		
		w3DContext(FALSE);
	}
	else
	{
		glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
		glMatrixMode(GL_MODELVIEW);
		glLoadIdentity();

		glShadeModel(GL_FLAT);
		glDisable(GL_LIGHTING);
		
		// apply camera transform
		pCamera->Apply();

		// render the scene
		w3DApplyAttrib(pModeler->GetAttrib());
		pModeler->SelectionRender();

		glShadeModel(GL_SMOOTH);
		glEnable(GL_LIGHTING);
	}
}

void w3DSystem::SetModeler(w3DModelBase *pModeler) 
{ 
	w3DSystem::pModeler = pModeler;
}

void w3DSystem::SetActiveCamera(w3DCamera *pCam) 
{ 
	pCamera = pCam; 
}

void w3DSystem::Stop() 
{ 
	stop = TRUE; 
}

void w3DSystem::Resume() 
{ 
	stop = FALSE; 
}

int w3DSystem::PickObject(int xPos, int yPos)
{
	static int selectBuff[256];
	int hits, viewport[4];
	unsigned long id;

	if (!w3DSystem::bInitialized || !w3DSystem::pModeler)
		return 0;

	w3DContext(TRUE);

	glGetIntegerv(GL_VIEWPORT, (int*)viewport);
	glSelectBuffer(256, (unsigned int*)selectBuff);

	glMatrixMode(GL_PROJECTION);
	glPushMatrix();

	glRenderMode(GL_SELECT);
	glInitNames();
	glPushName(0);

	glLoadIdentity();
	gluPickMatrix(xPos, (int)viewport[3] - yPos, 1, 1, (int*)viewport);

	gluPerspective(w3DSystem::pCamera->GetFOV(), w3DSystem::Aspect, 
					w3DSystem::pCamera->GetNearDist(), w3DSystem::pCamera->GetFarDist());

	w3DSystem::RenderingMode = FALSE;
	Process();
	w3DSystem::RenderingMode = TRUE;

	hits = glRenderMode(GL_RENDER); 

	if (hits <= 0)
	{
		glPopName();
		glMatrixMode(GL_PROJECTION);
		glPopMatrix();
		w3DContext(FALSE);
		return 0;
	}

	id = ProcessSelected(hits, (unsigned int*)selectBuff);

	glPopName();

	glMatrixMode(GL_PROJECTION);
	glPopMatrix();

	glFlush();

//	w3DDebugMsg("hits: %d, id: %d\n", hits, id);
	w3DContext(FALSE);

	return id;
}


unsigned long w3DSystem::ProcessSelected(GLint hits, GLuint *pSelectBuff)
{
	int i, j;
	GLuint names, *ptr;
	unsigned long id = 0;	// default: no object selected.
	unsigned long min = 0xFFFFFFFF;
	unsigned long tmp;
	

	ptr = (GLuint *)pSelectBuff;
	
	for (i = 0; i < (long)hits; i++) 
	{
		names = *ptr++;

		tmp = *ptr;

//		FDebugMsg("min: %u\n", *ptr);
		ptr++;

//		FDebugMsg("max: %u\n", *ptr);
		ptr++;
	
		if (min > tmp)
		{
			id = *ptr;
			min = tmp;
		}

		for (j = 0; j < names; j++) 
			*ptr++;
	}

	return id;
}




//////////////////////////////////////////////////////////////////////////////
// IMPLEMENTATION: w3DCamera

w3DCamera::w3DCamera(float x, float y, float z, float _pitch, float _yaw, float _roll) 
{
	Set(w3DPoint3(x, y, z), w3DPoint3(x, y, z - 1.0f), w3DVector3(0, 1, 0));
	Pitch(_pitch);
	Yaw(_yaw);
	Roll(_roll);
	
	ln = n;
	lu = u;
	lv = v;
	le = eye;
	
	fieldofview = 45.0f;
	nearDist = 15.0f;
	farDist = 10000.0f;
}

w3DCamera::w3DCamera()
{
	Set(w3DPoint3(), w3DPoint3(0, 0, -1), w3DVector3(0, 1, 0));
	
	ln = n;
	lu = u;
	lv = v;
	le = eye;

	fieldofview = 45.0f;
	nearDist = 15.0f;
	farDist = 10000.0f;
}

w3DCamera::~w3DCamera()
{}


void w3DCamera::Apply()
{
	static float m[16];

	m[0] = u.x;  m[4] = u.y;  m[8] = u.z;  m[12] = -eye.x;
	m[1] = v.x;  m[5] = v.y;  m[9] = v.z;  m[13] = -eye.y;
	m[2] = n.x;  m[6] = n.y;  m[10] = n.z; m[14] = -eye.z;
	m[3] = 0;    m[7] = 0;    m[11] = 0;   m[15] = 1.0;

	glMatrixMode(GL_MODELVIEW);
	glLoadMatrixf(m);
}

void w3DCamera::Set(w3DPoint3 e, w3DPoint3 l, w3DVector3 up)
{
	eye.Set(e);
	n.Set(eye.x - l.x, eye.y - l.y, eye.z - l.z);
	u.Set(up.Cross(n));
	n.Norm();
	u.Norm();
	v.Set(n.Cross(u));
}


void w3DCamera::Set(float x, float y, float z, float _pitch, float _yaw, float _roll) 
{
	Set(w3DPoint3(x, y, z), w3DPoint3(x, y, z - 1.0f), w3DVector3(0, 1, 0));
	Pitch(_pitch);
	Yaw(_yaw);
	Roll(_roll);
	
	ln = n;
	lu = u;
	lv = v;
	le = eye;
}

void w3DCamera::ResetAngle()
{
	n = ln;
	u = lu;
	v = lv;
}

void w3DCamera::Roll(float angle)
{
	float cs = (float)cos(-angle * w3DSystem::TORAD);
	float sn = (float)sin(-angle * w3DSystem::TORAD);

	w3DVector3 t = u;

	u.Set(cs*t.x - sn*v.x, cs*t.y - sn*v.y, cs*t.z - sn*v.z);
	v.Set(sn*t.x + cs*v.x, sn*t.y + cs*v.y, sn*t.z + cs*v.z);
}

void w3DCamera::Pitch(float angle)
{
	float cs = (float)cos(angle * w3DSystem::TORAD);
	float sn = (float)sin(angle * w3DSystem::TORAD);

	w3DVector3 t = n;

	n.Set(cs*t.x - sn*v.x, cs*t.y - sn*v.y, cs*t.z - sn*v.z);
	v.Set(sn*t.x + cs*v.x, sn*t.y + cs*v.y, sn*t.z + cs*v.z);
}

void w3DCamera::Yaw(float angle)
{
	float cs = (float)cos(angle * w3DSystem::TORAD);
	float sn = (float)sin(angle * w3DSystem::TORAD);

	w3DVector3 t = n;

	n.Set(cs*t.x - sn*u.x, cs*t.y - sn*u.y, cs*t.z - sn*u.z);
	u.Set(sn*t.x + cs*u.x, sn*t.y + cs*u.y, sn*t.z + cs*u.z);
}

void w3DCamera::Slide(float dU, float dV, float dN)
{
	eye.x += dU*u.x + dV*v.x + dN*n.x;
	eye.y += dU*u.y + dV*v.y + dN*n.y;
	eye.z += dU*u.z + dV*v.z + dN*n.z;
}

w3DPoint3 w3DCamera::GetPos()
{
	return eye;
}

void w3DCamera::SetParam(float vAng, float nearD, float farD)
{
	fieldofview = vAng;
	nearDist = nearD;
	farDist = farD;
}

float w3DCamera::GetFOV() 
{ 
	return fieldofview; 
}

float w3DCamera::GetNearDist() 
{ 
	return nearDist; 
}

float w3DCamera::GetFarDist() 
{ 
	return farDist; 
}



//////////////////////////////////////////////////////////////////////////////
// MISCELLANEOUS IMPLEMENTATIONS

void w3DContext(BOOL _ctx)
{
	if (!w3DSystem::bInitialized)
	{
#ifdef	W3DDEBUG
		w3DDebugMsg(W3D_ERR_NOTINITIALIZED);
#endif
		return;
	}

	if (_ctx)
		wglMakeCurrent(w3DSystem::hDC, w3DSystem::hRC);
	else
		wglMakeCurrent(NULL, NULL);
}

void w3DFlushAndSwap()
{
	w3DContext(TRUE);
	glFlush();
	SwapBuffers(w3DSystem::hDC);
	w3DContext(FALSE);
}

void w3DWinMsg(const char *fmt, ...)
{
	char		text[256];									// Holds Our String
	va_list		ap;											// Pointer To List Of Arguments

	if (fmt == NULL)										// If There's No Text
		return;												// Do Nothing

	va_start(ap, fmt);										// Parses The String For Variables
	    vsprintf(text, fmt, ap);							// And Converts Symbols To Actual Numbers
	va_end(ap);												// Results Are Stored In Text
	
	::MessageBox(NULL, text, "W3D System", MB_OK | MB_ICONERROR);
}

#pragma warning (disable : 4100)
void w3DDebugMsg(const char *fmt, ...)
{
	char		text[2560];									// Holds Our String
	va_list		ap;											// Pointer To List Of Arguments

	if (fmt == NULL)										// If There's No Text
		return;												// Do Nothing

	va_start(ap, fmt);										// Parses The String For Variables
	    vsprintf(text, fmt, ap);							// And Converts Symbols To Actual Numbers
	va_end(ap);												// Results Are Stored In Text

	OutputDebugString(text);
}
#pragma warning (default : 4100)


void w3DErrorTest(char* who)
{
	LONG i, error;
	
	for (i=0; i<6; i++)
	{
		error = glGetError();

		w3DDebugMsg("%s: %s\n", who, gluErrorString(error));

		if (error == GL_NO_ERROR)
			break;
	}
}

void w3DApplyAttrib(w3DGroup* pAttrGrp)
{
	// if there are no attributes...
	if (pAttrGrp == NULL)
		return;

	pAttrGrp->FindFirst();
	
	for (int i = 0; i < pAttrGrp->GetTotalNum(); i++)
	{
		if (classify_attrib(pAttrGrp->GetData()) == FALSE)
			break;
		pAttrGrp->FindNext();
	}
}

static BOOL classify_attrib(w3DObject *pObj)
{
	if (pObj == NULL)
		return FALSE;

	switch (pObj->GetType())
	{
	case W3D_ATTRIB_TRANSLATE:
		{
			static w3DTranslate *pTrans;
			pTrans = (w3DTranslate*)pObj;
			glTranslatef(pTrans->x, pTrans->y, pTrans->z);
		}
		break;

	case W3D_ATTRIB_ROTATE:
		{
			static w3DRotate *pRot;
			pRot = (w3DRotate*)pObj;
			glRotatef(pRot->a, pRot->x, pRot->y, pRot->z);
		}
		break;

	case W3D_ATTRIB_COLOR:
		{
			static w3DColor *pCol;
			pCol = (w3DColor*)pObj;
			glColor4fv(pCol->rgba);
		}
		break;
	}

	return TRUE;
}







//////////////////////////////////////////////////////////////////////////////
// IMPLEMENTATION: w3DModelRenderer

w3DModelRenderer::w3DModelRenderer()
{
	mtl = NULL;
	obj = NULL;
		
	cur_mtl = -1;
	cur_submtl = -1;
	cur_obj = -1;
	
	vtx_counter = 0;
	face_counter = 0;
	tvtx_counter = 0;
	tface_counter = 0;
}

w3DModelRenderer::~w3DModelRenderer()
{
}



void w3DModelRenderer::binReader()
{
	BFMSIGNATURE	sig;

	while (!bfm_bin.Err())
	{
		bfm_bin.ReadSig(&sig);

		switch(sig)
		{
		case SIG_NODENAME:
			{
				cur_obj++;
				ZeroMemory((void*)&obj[cur_obj], sizeof(STRUCT_OBJECT));
				bfm_bin.ReadInt8((PBFMINT8)obj[cur_obj].name, MAX_BF_NAME);

				// the object has no material by default.
				obj[cur_obj].mtlref = -1;
			}
			break;

		case SIG_NUM_VERTEX:
			{
				short v;
				bfm_bin.ReadInt16((PBFMINT16)&v, 1);
				obj[cur_obj].nVtx = v;
				obj[cur_obj].vtx = (STRUCT_VERTEX*)malloc(sizeof(STRUCT_VERTEX)*v);
				vtx_counter = 0;
			}
			break;

		case SIG_NUM_FACE:
			{
				short v;
				bfm_bin.ReadInt16((PBFMINT16)&v, 1);
				obj[cur_obj].nFace = v;
				obj[cur_obj].face = (STRUCT_FACE*)malloc(sizeof(STRUCT_FACE)*v);
				face_counter = 0;
			}
			break;

		case SIG_NUM_TVERTEX:
			{
				short v;
				bfm_bin.ReadInt16((PBFMINT16)&v, 1);
				obj[cur_obj].nTVtx = v;
				obj[cur_obj].tvtx = (STRUCT_TVERTEX*)malloc(sizeof(STRUCT_TVERTEX)*v);
				tvtx_counter = 0;
			}
			break;

		case SIG_NUM_TFACE:
			{
				short v;
				bfm_bin.ReadInt16((PBFMINT16)&v, 1);
				obj[cur_obj].nTFace = v;
				obj[cur_obj].tface = (STRUCT_TFACE*)malloc(sizeof(STRUCT_TFACE)*v);
				tface_counter = 0;
			}
			break;

		case SIG_WIREFRAMECOLOR:
			{
				bfm_bin.ReadFloat32((PBFMFLOAT32)obj[cur_obj].wfc, 3);
			}
			break;

		case SIG_VERTEX:
			{
				float f[3];
				bfm_bin.ReadFloat32((PBFMFLOAT32)f, 3);
				obj[cur_obj].vtx[vtx_counter].v[0] = f[0];
				obj[cur_obj].vtx[vtx_counter].v[1] = f[2];
				obj[cur_obj].vtx[vtx_counter].v[2] = -f[1];
				vtx_counter++;
			}
			break;

		case SIG_FACE:
			{
				short v[3];
				float vtx[3][3];

				bfm_bin.ReadInt16((PBFMINT16)v, 3);
				obj[cur_obj].face[face_counter].a = v[0];
				obj[cur_obj].face[face_counter].b = v[1];
				obj[cur_obj].face[face_counter].c = v[2];

				// face normal calculation.
				vtx[0][0] = obj[cur_obj].vtx[v[0]].v[0];
				vtx[0][1] = obj[cur_obj].vtx[v[0]].v[1];
				vtx[0][2] = obj[cur_obj].vtx[v[0]].v[2];
				
				vtx[1][0] = obj[cur_obj].vtx[v[1]].v[0];
				vtx[1][1] = obj[cur_obj].vtx[v[1]].v[1];
				vtx[1][2] = obj[cur_obj].vtx[v[1]].v[2];
				
				vtx[2][0] = obj[cur_obj].vtx[v[2]].v[0];
				vtx[2][1] = obj[cur_obj].vtx[v[2]].v[1];
				vtx[2][2] = obj[cur_obj].vtx[v[2]].v[2];
				
				calcNormalf(vtx, obj[cur_obj].face[face_counter].n);
				ReduceToUnitf(obj[cur_obj].face[face_counter].n);

				face_counter++;
			}
			break;

		case SIG_TVERTEX:
			{
				float f[2];

				// we ignore the z value
				bfm_bin.ReadFloat32((BFMFLOAT32*)f, 2);

				obj[cur_obj].tvtx[tvtx_counter].v[0] = f[0];
				obj[cur_obj].tvtx[tvtx_counter].v[1] = -f[1];
				tvtx_counter++;
			}
			break;

		case SIG_TFACE:
			{
				short l[3];
				
				bfm_bin.ReadInt16((BFMINT16*)l, 3);
				obj[cur_obj].tface[tface_counter].a = l[0];
				obj[cur_obj].tface[tface_counter].b = l[1];
				obj[cur_obj].tface[tface_counter].c = l[2];
				tface_counter++;
			}
			break;

		case SIG_MTLID:
			{
				short v;

				// read mtlid
				bfm_bin.ReadInt16((PBFMINT16)&v, 1);

				obj[cur_obj].face[face_counter-1].mtlid = v;
			}
			break;

		case SIG_MATERIAL:
			{
				short v;
				
				bfm_bin.ReadInt16((PBFMINT16)&v, 1);

				cur_mtl = v;
				cur_submtl = -1;	// no submtls by default.

				ZeroMemory((void*)&mtl[cur_mtl], sizeof(STRUCT_MATERIAL));
			}
			break;

		case SIG_MATERIAL_AMBIENT:
			{
				float f[3];
				
				bfm_bin.ReadFloat32((PBFMFLOAT32)f, 3);

				if (-1 != cur_submtl)
				{
					mtl[cur_mtl].submtl[cur_submtl].ambient[0] = f[0];
					mtl[cur_mtl].submtl[cur_submtl].ambient[1] = f[1];
					mtl[cur_mtl].submtl[cur_submtl].ambient[2] = f[2];
				}
				else
				{
					mtl[cur_mtl].ambient[0] = f[0];
					mtl[cur_mtl].ambient[1] = f[1];
					mtl[cur_mtl].ambient[2] = f[2];
				}
			}
			break;

		case SIG_MATERIAL_DIFFUSE:
			{
				float f[3];
				
				bfm_bin.ReadFloat32((PBFMFLOAT32)f, 3);

				if (-1 != cur_submtl)
				{
					mtl[cur_mtl].submtl[cur_submtl].diffuse[0] = f[0];
					mtl[cur_mtl].submtl[cur_submtl].diffuse[1] = f[1];
					mtl[cur_mtl].submtl[cur_submtl].diffuse[2] = f[2];
				}
				else
				{
					mtl[cur_mtl].diffuse[0] = f[0];
					mtl[cur_mtl].diffuse[1] = f[1];
					mtl[cur_mtl].diffuse[2] = f[2];
				}
			}
			break;

		case SIG_MATERIAL_SPECULAR:
			{
				float f[3];
				
				bfm_bin.ReadFloat32((PBFMFLOAT32)f, 3);

				if (-1 != cur_submtl)
				{
					mtl[cur_mtl].submtl[cur_submtl].specular[0] = f[0];
					mtl[cur_mtl].submtl[cur_submtl].specular[1] = f[1];
					mtl[cur_mtl].submtl[cur_submtl].specular[2] = f[2];
				}
				else
				{
					mtl[cur_mtl].specular[0] = f[0];
					mtl[cur_mtl].specular[1] = f[1];
					mtl[cur_mtl].specular[2] = f[2];
				}
			}
			break;

		case SIG_TEXTURE_NAME:
			{
				char name[32];
				BFMINT8 tid;
				long fs;
				char *buf;

				bfm_bin.ReadInt8((PBFMINT8)name, 32);
				bfm_bin.ReadInt8(&tid, 1);

				// read the texture size
				bfm_bin.ReadInt32((BFMINT32*)&fs, 1);

				if (fs > 0)
				{
					buf = (char*)malloc(fs);
					bfm_bin.ReadInt8((BFMINT8*)buf, fs);
				
					if (-1 != cur_submtl)
					{
						mtl[cur_mtl].submtl[cur_submtl].bTexture = TRUE;
						mtl[cur_mtl].submtl[cur_submtl].tid = tid;
					}
					else
					{
						// the material has a texture
						mtl[cur_mtl].bTexture = TRUE;
						mtl[cur_mtl].tid = tid;
					}
					
					tex[tid].LoadJPEG(buf, fs, GL_LINEAR, GL_REPEAT);
				}
			}
			break;

		case SIG_NUMSUBMTL:
			{
				short v;
				
				bfm_bin.ReadInt16((PBFMINT16)&v, 1);

				mtl[cur_mtl].submtl = (STRUCT_MATERIAL*)malloc(sizeof(STRUCT_MATERIAL)*v);
				ZeroMemory((void*)mtl[cur_mtl].submtl, sizeof(STRUCT_MATERIAL)*v);

				mtl[cur_mtl].nSubmtl = v;

				cur_submtl = -1;
			}
			break;

		case SIG_SUBMATERIAL:
			{
				short v;
				
				bfm_bin.ReadInt16((PBFMINT16)&v, 1);

				cur_submtl = v;
			}
			break;

		case SIG_MATERIAL_REF:
			{
				short v;
				
				// write object_count and mat_count
				bfm_bin.ReadInt16((PBFMINT16)&v, 1);

				obj[cur_obj].mtlref = v;
			}
			break;

		case SIG_GROUP:
			{
				char buf[32];
				
				ZeroMemory((void*)buf, 32);
				
				bfm_bin.ReadInt8((PBFMINT8)buf, 32);
			}
			break;

		case SIG_ENDGROUP:
			{

			}
			break;
		}

	}
}


void w3DModelRenderer::MakeNormals()
{
	register int	i, k;

#define MAX_NEIGHBOR	500
	typedef struct
	{
		int nFace[MAX_NEIGHBOR];
		int nCnt;
	} _neighbor;

	_neighbor*	pNeighbors;


	for (int objid=0; objid<fh.object_count; objid++)
	{
		// find neighboring faces
		pNeighbors = new _neighbor[obj[objid].nVtx];
		
		for (i=0; i<obj[objid].nVtx; i++)
		{
			for (k=0; k<MAX_NEIGHBOR; k++)
				pNeighbors[i].nFace[k] = -1;
			pNeighbors[i].nCnt = 0;
		}

		// count neighbors.
		for (i=0; i<obj[objid].nFace; i++)
		{
			int a = obj[objid].face[i].a;
			int b = obj[objid].face[i].b;
			int c = obj[objid].face[i].c;

			pNeighbors[a].nFace[pNeighbors[a].nCnt] = i;
			pNeighbors[a].nCnt++;

			pNeighbors[b].nFace[pNeighbors[b].nCnt] = i;
			pNeighbors[b].nCnt++;

			pNeighbors[c].nFace[pNeighbors[c].nCnt] = i;
			pNeighbors[c].nCnt++;
		}

		// calc vertex normals.
		for (i=0; i<obj[objid].nVtx; i++)
		{
			float acc[3] = {0.0f, 0.0f, 0.0f};

			for (k=0; k<pNeighbors[i].nCnt; k++)
			{
				int nf = pNeighbors[i].nFace[k];
				acc[0] += obj[objid].face[nf].n[0];
				acc[1] += obj[objid].face[nf].n[1];
				acc[2] += obj[objid].face[nf].n[2];
			}

			ReduceToUnitf(acc);

			obj[objid].vtx[i].n[0] = acc[0];
			obj[objid].vtx[i].n[1] = acc[1];
			obj[objid].vtx[i].n[2] = acc[2];
		}


		w3DVector3 tmp_max = w3DVector3(-FLT_MAX, -FLT_MAX, -FLT_MAX);
		w3DVector3 tmp_min = w3DVector3(FLT_MAX, FLT_MAX, FLT_MAX);
		// make bounding box for the object in the model.
		for (i=0; i<obj[objid].nVtx; i++)
		{
			float* pVtx = obj[objid].vtx[i].v;

			if (pVtx[0] < tmp_min.x) tmp_min.x = pVtx[0];
			if (pVtx[1] < tmp_min.y) tmp_min.y = pVtx[1];
			if (pVtx[2] < tmp_min.z) tmp_min.z = pVtx[2];

			if (pVtx[0] > tmp_max.x) tmp_max.x = pVtx[0];
			if (pVtx[1] > tmp_max.y) tmp_max.y = pVtx[1];
			if (pVtx[2] > tmp_max.z) tmp_max.z = pVtx[2];
		}

		obj[objid].min = tmp_min;
		obj[objid].max = tmp_max;
	}
}


void w3DModelRenderer::Init(char* fn)
{
	bfm_bin.Init(fn);
	bfm_bin.OpenToRead();

	bfm_bin.ReadInt8((PBFMINT8)&fh, sizeof(FILE_HEADER));

	obj = (STRUCT_OBJECT*)malloc(sizeof(STRUCT_OBJECT)*fh.object_count);
	mtl = (STRUCT_MATERIAL*)malloc(sizeof(STRUCT_MATERIAL)*fh.mat_count);

	binReader();
	MakeNormals();
}


void w3DModelRenderer::Render()
{
	int i;
	int k;
	int mtlref;
	int ta, tb, tc;
	int va, vb, vc;
	BOOL	bTexture;
	int		iSubmtl;


	for (k=0; k<fh.object_count; k++)
	{
		mtlref = obj[k].mtlref;

		for (i=0; i<obj[k].nFace; i++)
		{
			if (-1 == mtlref)
			{
				glColor3f(obj[k].wfc[0], obj[k].wfc[1], obj[k].wfc[2]);
			}
			else
			{
				if (mtl[mtlref].nSubmtl > 0)
				{	// this material has sub-material

					iSubmtl = obj[k].face[i].mtlid;

					if (iSubmtl > mtl[mtlref].nSubmtl)
						iSubmtl = mtl[mtlref].nSubmtl - 1;	// corrected.

					if (mtl[mtlref].submtl[iSubmtl].bTexture)
					{
						bTexture = TRUE;
						glEnable(GL_TEXTURE_2D);
						tex[mtl[mtlref].submtl[iSubmtl].tid].Apply();
					}
					else
						bTexture = FALSE;

					glMaterialfv(GL_FRONT, GL_AMBIENT, mtl[mtlref].submtl[iSubmtl].ambient);
					glMaterialfv(GL_FRONT, GL_DIFFUSE, mtl[mtlref].submtl[iSubmtl].diffuse);
					glMaterialfv(GL_FRONT, GL_SPECULAR, mtl[mtlref].submtl[iSubmtl].specular);
				}
				else
				{	// this material has no sub-material
					if (mtl[mtlref].bTexture)
					{
						bTexture = TRUE;
						glEnable(GL_TEXTURE_2D);
						tex[mtl[mtlref].tid].Apply();
					}
					else
						bTexture = FALSE;

					glMaterialfv(GL_FRONT, GL_AMBIENT, mtl[mtlref].ambient);
					glMaterialfv(GL_FRONT, GL_DIFFUSE, mtl[mtlref].diffuse);
					glMaterialfv(GL_FRONT, GL_SPECULAR, mtl[mtlref].specular);
				}
			}

			
			va = obj[k].face[i].a;
			vb = obj[k].face[i].b;
			vc = obj[k].face[i].c;
			
			if (!bTexture)
				glDisable(GL_TEXTURE_2D);

			
			glBegin(GL_TRIANGLES);
			
			if (bTexture)
			{
				ta = obj[k].tface[i].a;
				tb = obj[k].tface[i].b;
				tc = obj[k].tface[i].c;

				if (w3DSystem::SmoothShade)
				{
					glNormal3fv(obj[k].vtx[va].n);
					glTexCoord2fv(obj[k].tvtx[ta].v);
					glVertex3fv(obj[k].vtx[va].v);
					
					glNormal3fv(obj[k].vtx[vb].n);
					glTexCoord2fv(obj[k].tvtx[tb].v);
					glVertex3fv(obj[k].vtx[vb].v);
					
					glNormal3fv(obj[k].vtx[vc].n);
					glTexCoord2fv(obj[k].tvtx[tc].v);
					glVertex3fv(obj[k].vtx[vc].v);
				}
				else
				{
					glNormal3fv(obj[k].face[i].n);
					
					glTexCoord2fv(obj[k].tvtx[ta].v);
					glVertex3fv(obj[k].vtx[va].v);
					
					glTexCoord2fv(obj[k].tvtx[tb].v);
					glVertex3fv(obj[k].vtx[vb].v);
					
					glTexCoord2fv(obj[k].tvtx[tc].v);
					glVertex3fv(obj[k].vtx[vc].v);
				}
			}
			else
			{
				if (w3DSystem::SmoothShade)
				{
					glNormal3fv(obj[k].vtx[va].n);
					glVertex3fv(obj[k].vtx[va].v);
					glNormal3fv(obj[k].vtx[vb].n);
					glVertex3fv(obj[k].vtx[vb].v);
					glNormal3fv(obj[k].vtx[vc].n);
					glVertex3fv(obj[k].vtx[vc].v);
				}
				else
				{
					glNormal3fv(obj[k].face[i].n);
					
					glVertex3fv(obj[k].vtx[va].v);
					glVertex3fv(obj[k].vtx[vb].v);
					glVertex3fv(obj[k].vtx[vc].v);
				}
			}
			
			glEnd();
			
		}
	}
}

void w3DModelRenderer::SelectionRender()
{
	int i;
	int k;
	int va, vb, vc;

	for (k=0; k<fh.object_count; k++)
	{
		glLoadName(k+1);
		
		for (i=0; i<obj[k].nFace; i++)
		{
			va = obj[k].face[i].a;
			vb = obj[k].face[i].b;
			vc = obj[k].face[i].c;
			
			glBegin(GL_TRIANGLES);
			
			glVertex3fv(obj[k].vtx[va].v);
			glVertex3fv(obj[k].vtx[vb].v);
			glVertex3fv(obj[k].vtx[vc].v);
			
			glEnd();
		}
	}
}

void w3DModelRenderer::QueryObject(unsigned long id, void** anydata)
{
	unsigned long i;

	*anydata = NULL;

	for (i=0; i<fh.object_count; i++)
	{
		if (id == (i+1))
		{
			*anydata = (void*)&obj[i];
			break;
		}
	}
}

//////////////////////////////////////////////////////////////////////////////
// IMPLEMENTATION: w3DTexture

int		w3DTexture::texture_count = 0;
int		w3DTexture::texture_last = -1;
long	w3DTexture::texture_load_size = 0;


w3DTexture::w3DTexture()
{
	loadsize = 0;
	TID = -1;
}

w3DTexture::~w3DTexture()
{
	texture_count--;
	texture_load_size -= loadsize;
	loadsize = 0;

	w3DContext(TRUE);
	glDeleteTextures(1, (const unsigned int*)&TID);
	w3DContext(FALSE);
}

// INTERNAL FUNCTION
static BOOL _chkIJLError(int i)
{
	if (i == IJL_OK || i == IJL_INTERRUPT_OK || i == IJL_ROI_OK)
		return FALSE;

	w3DDebugMsg("IJL Error: %s\n", ijlErrorStr((IJLERR)i));

	return TRUE;
}

BOOL w3DTexture::LoadJPEG(char* buf, long size, int filter, int wrap, int tid)
{
    JPEG_CORE_PROPERTIES image;
	IJLIOTYPE opt_param, opt_read;
    char* imageData;
	long imgx, imgy;


    if (buf == NULL)
		return FALSE;

	ZeroMemory(&image, sizeof(JPEG_CORE_PROPERTIES));

    if(_chkIJLError(ijlInit( &image )))
    {
        ijlFree( &image );
		return 0;
    }


	if (0 == size)
	{
		opt_param = IJL_JFILE_READPARAMS;
		opt_read = IJL_JFILE_READWHOLEIMAGE;
		image.JPGFile = buf;
    }
	else
	{
		opt_param = IJL_JBUFF_READPARAMS;
		opt_read = IJL_JBUFF_READWHOLEIMAGE;
		image.JPGBytes = (unsigned char*)buf;
		image.JPGSizeBytes = size;
	}


    if(_chkIJLError(ijlRead( &image, opt_param )))
	{
		ijlFree( &image );
		return 0;
	}

    LONG nchannels = image.JPGChannels;
    LONG imageSize = image.JPGWidth * image.JPGHeight * nchannels;

    imageData = new char[imageSize];
	

    if(!imageData)
    {
        ijlFree( &image );
		return FALSE;
    }

    image.DIBWidth    = image.JPGWidth;
    image.DIBHeight   = image.JPGHeight;
    image.DIBChannels = nchannels;
    image.DIBBytes    = (unsigned char*)imageData;

    if (_chkIJLError(ijlRead( &image, opt_read )))
    {
        delete[] imageData;
        ijlFree( &image );
		return 0;
    }

	imgx = image.DIBWidth;
	imgy = image.DIBHeight;

	if (!chkPowerOf2(image.DIBWidth) || !chkPowerOf2(image.DIBHeight))
	{
		char* pScaledImage = NULL;
		short nx = getUpperPowerOf2(image.DIBWidth);
		short ny = getUpperPowerOf2(image.DIBHeight);
		if (3 == nchannels)
			FastZoom24(imageData, image.DIBWidth, image.DIBHeight, &pScaledImage, nx, ny);
		else if (1 == nchannels)
			FastZoom8(imageData, image.DIBWidth, image.DIBHeight, &pScaledImage, nx, ny);
		delete[] imageData;
		imageData = pScaledImage;
		imgx = nx;
		imgy = ny;
	}


	texture_load_size += imageSize;
	loadsize = imageSize;

	w3DContext(TRUE);

	if (-1 == tid)
	{
		glGenTextures(1, (unsigned int*)&TID);
	}
	else
	{
		TID = tid;
	}


	texture_last = TID;

    glBindTexture(GL_TEXTURE_2D, TID);
	glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER, filter);
	glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER, filter);
	glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_S, wrap);
	glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_T, wrap);
	glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);


	if (image.DIBColor == IJL_RGB)
		glTexImage2D(GL_TEXTURE_2D, 0, nchannels, imgx, imgy,
			0, GL_RGB, GL_UNSIGNED_BYTE, imageData);
	else
		glTexImage2D(GL_TEXTURE_2D, 0, nchannels, imgx, imgy,
			0, GL_BGR_EXT, GL_UNSIGNED_BYTE, imageData);

	delete[] imageData;
    ijlFree( &image );

	texture_count++;

	w3DContext(FALSE);

	return TRUE;
}


void w3DTexture::Apply()
{
	if (texture_last != TID)
	{
		glBindTexture(GL_TEXTURE_2D, TID);
		texture_last = TID;
	}
}

void w3DTexture::SetTID(int id)
{
	TID = id;
}

int w3DTexture::GetTID()
{
	return TID;
}

long w3DTexture::GetLoadSize()
{
	return loadsize;
}

long w3DTexture::GetTotalLoadSize()
{
	return texture_load_size;
}



#undef W3DDEBUG
