

/* Rev.2
** Binary File Manager, 2000-2001, Jinhong Kim, cheriz@empal.com
** Tested at Warning Level 4
*/


#include "bfm.h"
#include <windows.h>
#include <stdio.h>


BFMINT16 __strlen(char *str)
{
	BFMINT16 len = 0;
	while (*str++) len++;
	return len;
}


BFManager::BFManager()
{
	hFile = NULL;
	err = FALSE;
	file_size = 0;
	buf_ptr = NULL;
	buf_cur_ptr = NULL;
	buf_last_ptr = NULL;
	bDMA = FALSE;
}

BFManager::BFManager(char *fn)
{
	hFile = NULL;
	err = FALSE;
	file_size = 0;
	buf_ptr = NULL;
	buf_cur_ptr = NULL;
	buf_last_ptr = NULL;
	bDMA = FALSE;

	Init(fn);
}

BFManager::~BFManager()
{
	Close();
}

void BFManager::Close()
{
	if (buf_ptr && !bDMA)
		free(buf_ptr);

	if (hFile)
		CloseHandle(hFile);

	buf_ptr = NULL;
	buf_cur_ptr = NULL;
	buf_last_ptr = NULL;
}

void BFManager::Init(char *fname)
{
	BFMINT16 len = __strlen(fname);

	CopyMemory((void*)file_name, (void*)fname, len+1); 
}

void BFManager::InitDirectMemoryAccess(char* buf)
{
	buf_ptr = (PBFMINT8)buf;
	buf_cur_ptr = buf_ptr;
	bDMA = TRUE;
}

BFMINT16 BFManager::OpenToRead()
{
	hFile = CreateFile(file_name,
			GENERIC_READ,
			FILE_SHARE_READ,
			NULL,
			OPEN_EXISTING,
			FILE_ATTRIBUTE_NORMAL,
			NULL);
	
	if (hFile == INVALID_HANDLE_VALUE)
	{
		hFile = NULL;

		char buf[50];
		sprintf(buf, "%s not found!", file_name);
		::MessageBox(NULL, buf, "Error", MB_ICONEXCLAMATION);

		return BFMINT16_MAX;
	}

	file_size = GetFileSize(hFile, NULL);

	DWORD	read_size;
	buf_ptr = (BFMINT8*)malloc(file_size);
	buf_cur_ptr = buf_ptr;
	buf_last_ptr = buf_ptr + file_size - 1;

	ReadFile(hFile, (void*)buf_ptr, file_size, &read_size, NULL);

	return 0;
}


BFMINT16 BFManager::OpenToWrite()
{
	hFile = CreateFile(file_name,
			GENERIC_READ | GENERIC_WRITE,
			FILE_SHARE_READ,
			NULL,
			OPEN_ALWAYS,
			FILE_ATTRIBUTE_NORMAL,
			NULL);
	
	if (hFile == INVALID_HANDLE_VALUE)
	{
		hFile = NULL;

		char buf[50];
		sprintf(buf, "%s not found!", file_name);
		::MessageBox(NULL, buf, "Error", MB_ICONEXCLAMATION);

		return BFMINT16_MAX;
	}

	return 0;
}


// Read chunk from memory.

BFMINT16 BFManager::ReadSig(BFMSIGNATURE *v)
{
	*v = *buf_cur_ptr++;
		
	if (buf_cur_ptr > buf_last_ptr && !bDMA)
	{
		err = 1;
		return BFMINT16_MAX;
	}

	return 0;
}

BFMINT16 BFManager::ReadInt8(BFMINT8 *v, BFMINT32 len)
{
	BFMINT32 i;

	for (i=0; i<len; i++)
	{
		*(v+i) = *buf_cur_ptr++;
		if (buf_cur_ptr > buf_last_ptr && !bDMA)
		{
			err = 1;
			return BFMINT16_MAX;
		}
	}

	return 0;
}


BFMINT16 BFManager::ReadInt16(BFMINT16 *v, BFMINT32 len)
{
	BFMINT32 i;

	for (i=0; i<len; i++)
	{
		*(v+i) = *(BFMINT16*)buf_cur_ptr;
		buf_cur_ptr += sizeof(BFMINT16);

		if (buf_cur_ptr > buf_last_ptr && !bDMA)
		{
			err = 1;
			return BFMINT16_MAX;
		}
	}

	return 0;
}


BFMINT16 BFManager::ReadInt32(BFMINT32 *v, BFMINT32 len)
{
	BFMINT32 i;

	for (i=0; i<len; i++)
	{
		*(v+i) = *(BFMINT32*)buf_cur_ptr;
		buf_cur_ptr += sizeof(BFMINT32);

		if (buf_cur_ptr > buf_last_ptr && !bDMA)
		{
			err = 1;
			return BFMINT16_MAX;
		}
	}

	return 0;
}


BFMINT16 BFManager::ReadInt64(BFMINT64 *v, BFMINT32 len)
{
	BFMINT32 i;

	for (i=0; i<len; i++)
	{
		*(v+i) = *(BFMINT64*)buf_cur_ptr;
		buf_cur_ptr += sizeof(BFMINT64);

		if (buf_cur_ptr > buf_last_ptr && !bDMA)
		{
			err = 1;
			return BFMINT16_MAX;
		}
	}

	return 0;
}


BFMINT16 BFManager::ReadFloat32(BFMFLOAT32 *v, BFMINT32 len)
{
	BFMINT32 i;

	for (i=0; i<len; i++)
	{
		*(v+i) = *(BFMFLOAT32*)buf_cur_ptr;
		buf_cur_ptr += sizeof(BFMFLOAT32);

		if (buf_cur_ptr > buf_last_ptr && !bDMA)
		{
			err = 1;
			return BFMINT16_MAX;
		}
	}

	return 0;
}


BFMINT16 BFManager::ReadFloat64(BFMFLOAT64 *v, BFMINT32 len)
{
	BFMINT32 i;

	for (i=0; i<len; i++)
	{
		*(v+i) = *(BFMFLOAT64*)buf_cur_ptr;
		buf_cur_ptr += sizeof(BFMFLOAT64);

		if (buf_cur_ptr > buf_last_ptr && !bDMA)
		{
			err = 1;
			return BFMINT16_MAX;
		}
	}

	return 0;
}


BFMINT16 BFManager::ReadFloat80(BFMFLOAT80 *v, BFMINT32 len)
{
	BFMINT32 i;

	for (i=0; i<len; i++)
	{
		*(v+i) = *(BFMFLOAT80*)buf_cur_ptr;
		buf_cur_ptr += sizeof(BFMFLOAT80);

		if (buf_cur_ptr > buf_last_ptr && !bDMA)
		{
			err = 1;
			return BFMINT16_MAX;
		}
	}

	return 0;
}



BFMINT16 BFManager::WriteSig(BFMSIGNATURE v)
{
	DWORD written;
	WriteFile(hFile, (void*)&v, sizeof(BFMSIGNATURE), &written, NULL);

	return 0;
}


BFMINT16 BFManager::WriteInt8(BFMINT8 *v, BFMINT32 len)
{
	DWORD written;
	WriteFile(hFile, (void*)v, sizeof(BFMINT8) * len, &written, NULL);

	return 0;
}

BFMINT16 BFManager::WriteInt16(BFMINT16 *v, BFMINT32 len)
{
	DWORD written;
	WriteFile(hFile, (void*)v, sizeof(BFMINT16) * len, &written, NULL);

	return 0;
}

BFMINT16 BFManager::WriteInt32(BFMINT32 *v, BFMINT32 len)
{
	DWORD written;
	WriteFile(hFile, (void*)v, sizeof(BFMINT32) * len, &written, NULL);

	return 0;
}

BFMINT16 BFManager::WriteInt64(BFMINT64 *v, BFMINT32 len)
{
	DWORD written;
	WriteFile(hFile, (void*)v, sizeof(BFMINT64) * len, &written, NULL);

	return 0;
}


BFMINT16 BFManager::WriteFloat32(BFMFLOAT32 *v, BFMINT32 len)
{
	DWORD written;
	WriteFile(hFile, (void*)v, sizeof(BFMFLOAT32) * len, &written, NULL);

	return 0;
}

BFMINT16 BFManager::WriteFloat64(BFMFLOAT64 *v, BFMINT32 len)
{
	DWORD written;
	WriteFile(hFile, (void*)v, sizeof(BFMFLOAT64) * len, &written, NULL);

	return 0;
}

BFMINT16 BFManager::WriteFloat80(BFMFLOAT80 *v, BFMINT32 len)
{
	DWORD written;
	WriteFile(hFile, (void*)v, sizeof(BFMFLOAT80) * len, &written, NULL);

	return 0;
}


BFMINT16 BFManager::WriteSeek(BFMINT32 pos)
{
	SetFilePointer(hFile, pos, NULL, FILE_BEGIN);

	return 0;
}

BFMINT16 BFManager::WriteSkip(BFMINT32 pos)
{
	SetFilePointer(hFile, pos, NULL, FILE_CURRENT);

	return 0;
}

BFMINT16 BFManager::WriteCurrent(BFMINT32 *pos)
{
	*pos = SetFilePointer(hFile, 0, NULL, FILE_CURRENT);

	return 0;
}

BFMINT16 BFManager::WriteRewind()
{
	SetFilePointer(hFile, 0, NULL, FILE_BEGIN);

	return 0;
}


BFMINT16 BFManager::Seek(BFMINT32 pos)
{
	buf_cur_ptr = buf_ptr + pos;

	return 0;
}

BFMINT16 BFManager::Skip(BFMINT32 len)
{
	buf_cur_ptr += len;

	return 0;
}

BFMINT16 BFManager::Current(BFMINT32 *pos)
{
	*pos = buf_cur_ptr - buf_ptr;

	return 0;
}

BFMINT16 BFManager::Rewind()
{
	buf_cur_ptr = buf_ptr;

	return 0;
}


BFMINT16 BFManager::Err()
{
	return err;
}

BFMINT32 BFManager::GetFSize()
{
	return file_size;
}

