


// Define 'W3DDEBUG' to watch the debug string output.

// START OF W3D


#ifndef _W3D_DEFINITION_
#define _W3D_DEFINITION_


#pragma warning(disable: 4251)

#ifdef W3D_EXPORTS
#define W3D_API __declspec(dllexport)
#else
#define W3D_API __declspec(dllimport)
#endif



#include <windows.h>
#include <gl/gl.h>
#include <gl/glu.h>
#include <gl/glaux.h>
#include <math.h>
#include "w3dmath.h"
#include "dbllist.h"
#include "bfm.h"
#include "jpg/ijl.h"
#include "brsz.h"
#include <stdlib.h>


//////////////////////////////////////////////////////////////////////////////
// W3D CLASS DEFINITIONS


#define W3D_ERR_OK				0x00000000
#define W3D_ERR_FAIL			0x00000001

#define W3D_CLS_OBJECT			0x10000001
#define W3D_CLS_ATTRIB			0x10000002
#define W3D_CLS_MODEL			0x10000003
#define W3D_CLS_SYSTEM			0x10000004
#define W3D_CLS_CAMERA			0x10000005
#define W3D_CLS_GROUP			0x10000006

#define	W3D_ATTRIB_TRANSLATE	0x20000001
#define	W3D_ATTRIB_COLOR		0x20000002
#define	W3D_ATTRIB_ROTATE		0x20000003


#define DECLARE_W3D_CLASS_TYPE(a)		virtual DWORD GetType() { return a; }

#define NAMESIZE	32




#define NUM_VERTEX			"*MESH_NUMVERTEX"
#define NUM_FACE			"*MESH_NUMFACES"
#define NUM_TVERTEX			"*MESH_NUMTVERTEX"
#define NUM_TFACE			"*MESH_NUMTVFACES"
#define VERTEX_LIST			"*MESH_VERTEX_LIST"
#define VERTEX				"*MESH_VERTEX"
#define FACE_LIST			"*MESH_FACE_LIST"
#define FACE				"*MESH_FACE"
#define NORMALS				"*MESH_NORMALS"
#define FACE_NORMAL			"*MESH_FACENORMAL"
#define NVERTEX				"*MESH_VERTEXNORMAL"
#define TVERTEX				"*MESH_TVERT"
#define TFACE				"*MESH_TFACE"
#define TEXTURE				"*BITMAP"
#define UTILE				"*UVW_U_TILING"
#define VTILE				"*UVW_V_TILING"
#define UOFFSET				"*UVW_U_OFFSET"
#define VOFFSET				"*UVW_V_OFFSET"
#define GEOMOBJECT			"*GEOMOBJECT"
#define NODENAME			"*NODE_NAME"
#define WIREFRAMECOLOR		"*WIREFRAME_COLOR"
#define GROUP				"*GROUP"
#define MAT_COUNT			"*MATERIAL_COUNT"
#define MAT_MATERIAL		"*MATERIAL"
#define MAT_REFERENCE		"*MATERIAL_REF"
#define MAT_SUBMATERIAL		"*SUBMATERIAL"
#define MAT_NUMSUBMTL		"*NUMSUBMTLS"
#define MAT_AMBIENT			"*MATERIAL_AMBIENT"
#define MAT_DIFFUSE			"*MATERIAL_DIFFUSE"
#define MAT_SPECULAR		"*MATERIAL_SPECULAR"
#define MESH_MTLID			"*MESH_MTLID"



#define SIG_MATERIAL_COUNT			(BFMINT8)1
#define SIG_MATERIAL				(BFMINT8)2
#define SIG_MATERIAL_AMBIENT		(BFMINT8)3
#define SIG_MATERIAL_DIFFUSE		(BFMINT8)4
#define SIG_MATERIAL_SPECULAR		(BFMINT8)5
#define SIG_TEXTURE_NAME			(BFMINT8)6
#define SIG_NODENAME				(BFMINT8)7
#define SIG_WIREFRAMECOLOR			(BFMINT8)8
#define SIG_TEXTURE_UTILE			(BFMINT8)9
#define SIG_TEXTURE_VTILE			(BFMINT8)10
#define SIG_TEXTURE_UOFFSET			(BFMINT8)11
#define SIG_TEXTURE_VOFFSET			(BFMINT8)12
#define SIG_GEOMOBJECT				(BFMINT8)13
#define SIG_NUM_VERTEX				(BFMINT8)14
#define SIG_NUM_FACE				(BFMINT8)15
#define SIG_NUM_TVERTEX				(BFMINT8)16
#define SIG_NUM_TFACE				(BFMINT8)17
#define SIG_VERTEX					(BFMINT8)18
#define SIG_FACE					(BFMINT8)19
#define SIG_TVERTEX					(BFMINT8)20
#define SIG_TFACE					(BFMINT8)21
#define SIG_MATERIAL_REF			(BFMINT8)22
#define SIG_SUBMATERIAL				(BFMINT8)23
#define SIG_NUMSUBMTL				(BFMINT8)24
#define SIG_GROUP					(BFMINT8)25
#define SIG_ENDGROUP				(BFMINT8)26
#define SIG_TEXTURE_RES				(BFMINT8)27
#define SIG_MTLID					(BFMINT8)28


typedef struct
{
	BFMINT8	object_count;
	BFMINT8	mat_count;
	BFMINT16 face_count;
} FILE_HEADER;

typedef struct
{
	float v[3];
	float n[3];		// vertex normal
} STRUCT_VERTEX;

typedef struct
{
	int a;
	int b;
	int c;
	float n[3];		// face normal
	int mtlid;
} STRUCT_FACE;

typedef struct
{
	float v[2];
} STRUCT_TVERTEX;

typedef struct
{
	int a;
	int b;
	int c;
} STRUCT_TFACE;

typedef struct _STRUCT_MATERIAL
{
	float ambient[4];
	float diffuse[4];
	float specular[4];

	struct _STRUCT_MATERIAL* submtl;
	int nSubmtl;
	BOOL bTexture;

	int tid;
} STRUCT_MATERIAL;


typedef struct
{
	char name[32];

	int nVtx;
	int nFace;
	int nTVtx;
	int nTFace;

	float wfc[3]; // wire frame color

	STRUCT_VERTEX* vtx;
	STRUCT_FACE* face;
	STRUCT_TVERTEX* tvtx;
	STRUCT_TFACE* tface;

	int mtlref;
	w3DVector3	min, max;
} STRUCT_OBJECT;




//////////////////////////////////////////////////////////////////////////////
// W3D CLASS FORWARD DECLARATIONS

class w3DObject;
class w3DCamera;
class w3DLoadIdentity;
class w3DTranslate;
class w3DRotate;
class w3DColor;
class w3DModelBase;
class w3DSystem;
class w3DGroup;
class w3DTexture;
class w3DModelRenderer;


//////////////////////////////////////////////////////////////////////////////
// W3D HELPER FUNCTIONS PROTOTYPE FORWARD DECLARATIONS

void W3D_API w3DContext(BOOL _ctx);
void W3D_API w3DFlushAndSwap();
void W3D_API w3DErrorTest(char *who);
void W3D_API w3DWinMsg(const char *fmt, ...);
void W3D_API w3DDebugMsg(const char *fmt, ...);
void W3D_API w3DApplyAttrib(w3DGroup *pGrp);


//////////////////////////////////////////////////////////////////////////////
// DECLARATION: w3DObject

class W3D_API w3DObject
{
public:
	w3DObject();
	~w3DObject();

	DECLARE_W3D_CLASS_TYPE(W3D_CLS_OBJECT)
};


//////////////////////////////////////////////////////////////////////////////
// DECLARATION: w3DGroup

typedef LinkedList<w3DObject*> LISTW3D;

#define W3DGROUPNODETYPE	LISTW3D::Node

class W3D_API w3DGroup : public w3DObject
{
private:
	LISTW3D					list;
	W3DGROUPNODETYPE		node;
	w3DGroup*				pParent;

public:
	w3DGroup();
	~w3DGroup();

	W3DGROUPNODETYPE* Add(w3DObject* item);
	void FindNext();
	void FindFirst();
	w3DObject* GetData();
	W3DGROUPNODETYPE& GetCurNode();
	void SetCurNode(W3DGROUPNODETYPE* pnode);
	BOOL DeleteFirst();
	void Empty();
	void SetParent(w3DGroup* pParent_);
	w3DGroup* GetParent();
	int GetTotalNum();

	DECLARE_W3D_CLASS_TYPE(W3D_CLS_GROUP)
};



//////////////////////////////////////////////////////////////////////////////
// DECLARATION: w3DCamera

class W3D_API w3DCamera : public w3DObject
{
private:
	float fieldofview, nearDist, farDist;
	w3DVector3 lu, lv, ln;
	w3DPoint3 le;

public:

	w3DPoint3 eye;
	w3DVector3 u, v, n;

	w3DCamera(float x, float y, float z, float _pitch, float _yaw, float _roll);
	w3DCamera();
	~w3DCamera();

	void Set(w3DPoint3 e, w3DPoint3 l, w3DVector3 up);   // eye, look, up
	void Set(float x, float y, float z, float _pitch, float _yaw, float _roll);
	void ResetAngle();
	void Yaw(float angle);
	void Pitch(float angle);
	void Roll(float angle);
	void Slide(float dU, float dV, float dN);

	void SetParam(float vAng, float nearD, float farD);
	float GetFOV();
	float GetNearDist();
	float GetFarDist();

	w3DPoint3 GetPos();
	float GetRoll();
	float GetYaw();

	void Apply();

	DECLARE_W3D_CLASS_TYPE(W3D_CLS_CAMERA)
};



//////////////////////////////////////////////////////////////////////////////
// DECLARATION & IMPLEMENTATION: w3DTranslate

class W3D_API w3DTranslate : public w3DObject
{
public:
	GLfloat x, y, z;

public:
	w3DTranslate(GLfloat x, GLfloat y, GLfloat z)
	{
		this->x = x;
		this->y = y;
		this->z = z;
	}
	w3DTranslate() : x(0), y(0), z(0) {}
	~w3DTranslate() {}

	void Set(GLfloat x, GLfloat y, GLfloat z)
	{
		this->x = x;
		this->y = y;
		this->z = z;
	}

	DECLARE_W3D_CLASS_TYPE(W3D_ATTRIB_TRANSLATE)
};



//////////////////////////////////////////////////////////////////////////////
// DECLARATION & IMPLEMENTATION: w3DRotate

class W3D_API w3DRotate : public w3DObject
{
public:
	GLfloat a, x, y, z;

public:
	w3DRotate(GLfloat a, GLfloat x, GLfloat y, GLfloat z)
	{
		this->a = a;
		this->x = x;
		this->y = y;
		this->z = z;
	}
	w3DRotate() : a(0), x(0), y(0), z(0) {}
	~w3DRotate() {}

	void Set(GLfloat a, GLfloat x, GLfloat y, GLfloat z)
	{
		this->a = a;
		this->x = x;
		this->y = y;
		this->z = z;
	}

	void SetAngle(GLfloat a)
	{
		this->a = a;
	}

	DECLARE_W3D_CLASS_TYPE(W3D_ATTRIB_ROTATE)
};


//////////////////////////////////////////////////////////////////////////////
// DECLARATION & IMPLEMENTATION: w3DColor

class W3D_API w3DColor : public w3DObject
{
public:
	GLfloat rgba[4];

public:
	w3DColor(GLfloat rgba[4])
	{
		this->rgba[0] = rgba[0];
		this->rgba[1] = rgba[1];
		this->rgba[2] = rgba[2];
		this->rgba[3] = rgba[3];
	}
	w3DColor(GLfloat r, GLfloat g, GLfloat b, GLfloat a)
	{
		this->rgba[0] = r;
		this->rgba[1] = g;
		this->rgba[2] = b;
		this->rgba[3] = a;
	}
	w3DColor() 
	{
		rgba[0] = rgba[1] = rgba[2] = rgba[3] = 0.0f;
	}
	~w3DColor() {}

	void Set(GLfloat rgba[4])
	{
		this->rgba[0] = rgba[0];
		this->rgba[1] = rgba[1];
		this->rgba[2] = rgba[2];
		this->rgba[3] = rgba[3];
	}
	void Set(GLfloat r, GLfloat g, GLfloat b, GLfloat a)
	{
		this->rgba[0] = r;
		this->rgba[1] = g;
		this->rgba[2] = b;
		this->rgba[3] = a;
	}

	DECLARE_W3D_CLASS_TYPE(W3D_ATTRIB_COLOR)
};



//////////////////////////////////////////////////////////////////////////////
// DECLARATION: w3DModel

class W3D_API w3DModelBase : public w3DObject
{
private:
	w3DGroup* attr;

public:
	w3DModelBase();
	~w3DModelBase();
	virtual void Render() = 0;			// render the model
	virtual void SelectionRender() {}
	virtual void QueryObject(unsigned long id, void** anydata) { *anydata = NULL; }

	void SetAttrib(w3DGroup *attr);
	w3DGroup* GetAttrib();

	DECLARE_W3D_CLASS_TYPE(W3D_CLS_MODEL)
};


//////////////////////////////////////////////////////////////////////////////
// DECLARATION: w3DSystem

class W3D_API w3DSystem : public w3DObject
{
private:
	BOOL stop;

	void SetupRC();
	void RenderScene();
	unsigned long ProcessSelected(GLint hits, GLuint *pSelectBuff);

public:

	static w3DModelBase* pModeler;
	static const GLfloat PI;
	static const GLfloat TORAD;
	static const GLfloat TODEG;
	static GLfloat		Aspect;
	static GLfloat		bkColor[4];
	static BOOL			cull_face;
	static BOOL			PolygonMode;
	static BOOL			SmoothShade;
	static BOOL			RenderingMode;	// TRUE: Normal, FALSE: Selection
	static GLfloat		ambientLight[4];
	static GLfloat		diffuseLight[4];
	static GLfloat		specularLight[4];
	static GLfloat		lightPos[4];
	static BYTE			DepthBufferSize;
	static BYTE			ColorDepth;

	static HWND			hWnd;
	static HDC			hDC;
	static HGLRC		hRC;
	static BOOL			bInitialized;
	static w3DCamera*	pCamera;


	w3DSystem();
	~w3DSystem();

	void Init(HWND hWnd);
	void AdjustCSInfo(CREATESTRUCT &cs);

	void SetModeler(w3DModelBase *pModeler);
	void SetActiveCamera(w3DCamera* pCam);
	void Resize(GLsizei w, GLsizei h);
	void Process();
	void Stop();
	void Resume();
	int PickObject(int xPos, int yPos);
	
	static void SetAmbientLight(GLfloat r, GLfloat g, GLfloat b, GLfloat a);
	static void SetDiffuseLight(GLfloat r, GLfloat g, GLfloat b, GLfloat a);
	static void SetSpecularLight(GLfloat r, GLfloat g, GLfloat b, GLfloat a);
	static void SetLightPos(GLfloat x, GLfloat y, GLfloat z, GLfloat a);
	static void SetBackGroundColor(GLfloat r, GLfloat g, GLfloat b, GLfloat a);
	static void SetCullFace(BOOL bi);
	static void SetPolygonMode(BOOL mode);
	static void SetColorDepth(BYTE cd);
	static void SetDepthBufferSize(BYTE db);


	DECLARE_W3D_CLASS_TYPE(W3D_CLS_SYSTEM)
};


//////////////////////////////////////////////////////////////////////////////
// DECLARATION: w3DTexture

class W3D_API w3DTexture
{
public:
	long loadsize;

	w3DTexture();
	~w3DTexture();

	BOOL LoadJPEG(char* buf, long size, int filter, int wrap, int tid = -1);
	void Apply();
	int GetTID();
	void SetTID(int id);
	long GetLoadSize();

	static long GetTotalLoadSize();

private:
	int	TID;

	static	int	texture_count;
	static  int	texture_last;
	static	long	texture_load_size;
};




//////////////////////////////////////////////////////////////////////////////
// DECLARATION: w3DModelRenderer

class W3D_API w3DModelRenderer : public w3DModelBase
{
private:
	STRUCT_MATERIAL*	mtl;
	STRUCT_OBJECT*		obj;
	FILE_HEADER			fh;
	BFManager			bfm_bin;
	w3DTexture			tex[256];
	w3DVector3			min_all, max_all;
	
	int cur_mtl;
	int cur_submtl;
	int cur_obj;
	
	int vtx_counter;
	int face_counter;
	int tvtx_counter;
	int tface_counter;

	void Render();
	void SelectionRender();

	void binReader();
	void MakeNormals();

public:
	w3DModelRenderer();
	~w3DModelRenderer();

	void Init(char* fn);
	void QueryObject(unsigned long id, void** anydata);

	DECLARE_W3D_CLASS_TYPE(W3D_CLS_MODEL)
};


#endif
